import json

from IPython import get_ipython
from IPython.display import display, Javascript
from importlib.resources import read_text

from .utils import notebook_path
from .states import quick_save, quick_load


def get_cell_index_for_code(cell_code):
    path = notebook_path()
    if path is None:
        raise RuntimeError("Cannot find notebook path")

    with open(path, 'r', encoding='utf-8') as f:
        nb = json.load(f)

    for cell in nb['cells']:
        if cell['cell_type'] == 'code':
            source = "".join(cell['source'])
            if source.strip() == cell_code.strip():
                return cell.get('metadata', {}).get('cell_index')
    return None


# --------------------------------------------

def pre_run_cell(info):
    """Runs before each cell, performs quick_load"""
    cell_code = info.raw_cell  # actual code
    index = get_cell_index_for_code(cell_code) - 1
    if index >= 0:  # It should be, but just to be sure
        # print(f"Loading checkpoint #{index}")  # For debugging
        quick_load(index=index)


def post_run_cell(result):
    """Runs after each cell, performs quick_save"""
    if hasattr(result, "info"):
        cell_code = result.info.raw_cell
    else:  # Get last cell from IPython history
        shell = get_ipython()
        cell_code = shell.history_manager.input_hist_raw[result.execution_count]

    index = get_cell_index_for_code(cell_code)
    # print(f"Saving checkpoint #{index}")  # For debugging
    quick_save(index=index)


# --------------------------------------------

def load_ipython_extension(ipython):
    """Tells IPython to load JS extension and perform before-cell and after-cell actions"""
    # add_cell_index_to_cells()  # Annotate the notebook before anything else
    _load_js_extension()
    ipython.events.register('pre_run_cell', pre_run_cell)
    ipython.events.register('post_run_cell', post_run_cell)


def unload_ipython_extension(ipython):
    """Clean up IPython"""
    ipython.events.unregister('pre_run_cell', pre_run_cell)
    ipython.events.unregister('post_run_cell', post_run_cell)


# ---------------------------------

def _load_js_extension():
    """Load and run cell indexing extension"""
    js_code = read_text('immersim.js_extensions', 'nb_cell_indexing.js')  # Load Extension
    display(Javascript(js_code))  # Run that extension
    # print("Cell indexing JS extension loaded!")



import sys
from types import ModuleType

import dill

from .utils import notebook_name, filepath, ISType, saved_vars_blacklist, saved_module_name

# -------------------------------------------------------------------
# Using dill (saving/loading the entire session)
def quick_save(index=None):
    """Save all user vars and modules in their current state

       Will not save variables beginning with _

       Saves modules by reference, not value"""
    nb_name = notebook_name()

    # Old Method (no filter, saves all metadata, which can screw up loading)
    # dill.dump_module(save_directory+ nb_name +pickle_extension, refimported=True)

    # New Method (can filter, saves new module with only useful variables)
    main_module = sys.modules['__main__']
    filtered_namespace = {var: value for var, value in main_module.__dict__.items()
                          if not var.startswith('_') and var not in saved_vars_blacklist}

    # Create a new module with filtered namespace
    saved_module = ModuleType(saved_module_name)
    saved_module.__dict__.update(filtered_namespace)

    # pprint(filtered_namespace)

    dill.dump_module(filepath(nb_name, ISType.PICKLE, index=index), saved_module,  refimported=True)


def quick_load(index=None):
    """Load all user vars and modules from a quick_save

       Will overwrite variables with the same name as saved variables

       Will not remove current variables that were not part of the save

       Example:
           x = 5; quick_save(); y = 9; quick_load(); # namespace is x=5, y=9"""
    nb_name = notebook_name()

    # Old Method
    # dill.load_module(save_directory+ nb_name +pickle_extension)

    # New Method (load named module to update main module from)
    loaded_module = dill.load_module(filepath(nb_name, ISType.PICKLE, index=index))

    # Copy saved variables into __main__ (doesn't copy save module metadata, which includes wrong __name__, etc.)
    main_namespace = sys.modules['__main__'].__dict__
    main_namespace.update({var: value for var, value in loaded_module.__dict__.items() if not var.startswith('__')})


# --- Fun Aliases ---
F5 = quick_save
F9 = quick_load



from .items import store, get_shelf_dict, open_shelf, retrieve, clear_shelf
from .states import quick_save, quick_load, F5, F9
# Do not import enable_checkpoints here, since importing that enables checkpoints


import inspect

from .utils import notebook_name, filepath, ISType, open_shelf_with_cleanup

# -------------------------------------------------------------------
# Using shelve to easily load and store specific items
# Idea: Each notebook has two associated shelves, a "temp" shelf and a "permanent" shelf
# The "temp" shelf could be reset near the start of the file, to contain results from last run experiment
#   or alternatively, some sort of stack of results that haven't been saved or discarded yet

def store(*items):
    """Store one (or more) items on this notebook's shelf

       Expects variables passed in (not just values)"""
    nb_name = notebook_name()

    # Get caller's local variables
    caller_locals = inspect.currentframe().f_back.f_locals

    # Derive caller's variable names for the requested items
    item_dict = {}
    for item in items:
        # match item to name: item pair
        names = [name for name, value in caller_locals.items() if value is item and not name.startswith('_')]
        if not names:  # Doesn't exist
            raise ValueError(f'Could not find variable name for object: {item}')
        print(f'names: {names}')
        item_dict[names[0]] = item

    # Put onto shelf
    with open_shelf_with_cleanup(filepath(nb_name, ISType.SHELVE), flag='c') as shelf:
        for name, item in item_dict.items():
            shelf[name] = item

    return


def get_shelf_dict():
    """Return the shelf dict associated with calling file."""
    nb_name = notebook_name()
    with open_shelf_with_cleanup(filepath(nb_name, ISType.SHELVE), flag='r') as shelf:
        shelf_dict = dict(shelf)
    return shelf_dict


def open_shelf(flag='c', directory=None, filename=notebook_name()):
    """Return the opened shelf associated with calling file, unless another filename/directory specified.
       Should be used in "with" statement.

       Example:
           with open_shelf() as my_shelf:
               my_shelf['lighting'] = 'mood'"""
    kwargs = {'directory': directory} if directory is not None else {}  # Conditionally override directory
    return open_shelf_with_cleanup(filepath(filename, ISType.SHELVE, **kwargs), flag=flag)


def retrieve(*items, directory=None, filename=notebook_name()):
    """Retrieve one (or more) items from shelf associated with calling file,
       unless another filename/directory specified."""
    kwargs = {'directory': directory} if directory is not None else {}  # Conditionally override directory
    with open_shelf_with_cleanup(filepath(filename, ISType.SHELVE, **kwargs), flag='r') as shelf:
        shelf_dict = dict(shelf)

    retrieved_items = []
    for item in items:
        if item not in shelf_dict:  # Doesn't exist
            raise ValueError(f'Could not find stored item: {item}')
        retrieved_items.append(shelf_dict[item])

    return tuple(retrieved_items)


def clear_shelf(*items, directory=None, filename=notebook_name()):
    """Delete one (or more) items from shelf associated with calling file,
       unless another filename/directory specified.

       By default, calling with no items passed will clear ALL items from shelf.

       Example:
           clear_shelf()

           clear_shelf(var1, var2)

           clear_shelf(filename='workbench')"""
    # Get caller's local variables
    caller_locals = inspect.currentframe().f_back.f_locals

    kwargs = {'directory': directory} if directory is not None else {}  # Conditionally override directory

    with open_shelf_with_cleanup(filepath(filename, ISType.SHELVE, **kwargs), flag='c') as shelf:
        if items:
            # Derive caller's variable names for the requested items
            item_list = []
            for item in items:
                # match item to name: item pair
                names = [name for name, value in caller_locals.items() if value is item and not name.startswith('_')]
                if not names:  # Doesn't exist
                    raise ValueError(f'Could not find variable name for object: {item}')
                item_list.append(names[0])
        else:
            item_list = list(shelf.keys())

        for item in item_list:
            del shelf[item]
    return



import os
import shelve
import sqlite3
from enum import Enum

from ipynbname import name as notebook_name
from ipynbname import path as notebook_path


warehouse_directory = r'warehouse'
pickle_extension = r'.pkl'
shelf_extension = r'.shelf'
saved_module_name = 'save_state'  # Name given to modules that are saved and loaded with this package
saved_vars_blacklist = ['In', 'Out', 'open', 'exit', 'quit', 'get_ipython', 'ipykernal',  # IPython vars
                        'shell', 'InteractiveShell',  # Command-Line vars
                        'pydev_jupyter_vars', 'remove_imported_pydev_package',  # Pycharm vars
                        'debugpy', 'ptvsd']  # VSCode vars

class ISType(Enum):
    PICKLE = "pickle"
    SHELVE = "shelve"

def ensure_directory(directory=warehouse_directory, filename=None, checkpoints=False):
    """Ensure all necessary directories exist, creating if necessary

       Also creates README.md for warehouse directory"""
    if filename:  # Ensure directory for a notebook
        os.makedirs(os.path.join(directory, filename), exist_ok=True)

        if checkpoints:  # Also ensure directory for checkpoints
            os.makedirs(os.path.join(directory, filename, '.checkpoints'), exist_ok=True)

    else:  # Ensure warehouse directory (probably only run at import time)
        os.makedirs(directory, exist_ok=True)
        with open(os.path.join(directory, 'README.md'), 'w') as label_file:
            label_file.write(f'This directory (/{warehouse_directory}) '
                             f'provides storage for the Immersive Simulations package')

ensure_directory()


def filepath(filename: str, method: ISType, directory: str=warehouse_directory, index: int=None) -> str:
    """Returns the exact filepath for a specific file and operation"""
    match method:  # Get correct extension
        case ISType.PICKLE:  # Quick-Save or Checkpoint
            if index is not None:  # Used for checkpoints
                save_filename = f'.checkpoints/checkpoint_{index}{pickle_extension}'
            else:  # Used for quick-save
                save_filename = f'quicksave{pickle_extension}'

        case ISType.SHELVE:  # Shelf
            save_filename = f'storage{shelf_extension}'

        case _:
            raise ValueError(f'Invalid method: {method}')

    # Ensure room in the warehouse
    ensure_directory(directory=directory, filename=filename, checkpoints=index is not None)

    return os.path.join(directory, filename, save_filename)  # Ex. warehouse/workbench/quicksave.pkl


class ShelfWithCleanup:
    """A shelf wrapper class that removes -wal and -shm objects upon close"""
    def __init__(self, path, flag='c'):
        self.path = path
        self.flag = flag
        self.shelf = None

    def __enter__(self):
        self.shelf = shelve.open(self.path, self.flag)
        return self.shelf

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.shelf.close()
        # Clean up SQLite WAL/SHM files if they exist
        for ext in ('-wal', '-shm'):
            try:
                os.remove(self.path + ext)
            except FileNotFoundError:
                pass
        return False  # don't suppress exceptions

def open_shelf_with_cleanup(path, flag='c'):
    """Opens and returns a ShelfWithCleanup, which is a shelf that deletes temporary files to clean up after use"""
    return ShelfWithCleanup(path, flag)
